/***************************************************************************
*                         Fast Correlation Attack Implementation
*                          Meier & Staffelbach's Algorithm B
*
*   File    : algorithm_b.c
*   Purpose : Provides function to solve an instance of the correlation
*             problem using Meier & Staffelbach's algorithm B
*             
*
*   Example: if we have a correlation with probablility p
*             to and lfsr, of length l with the t taps, taps[]:
*
*             algorithm_b(p, l, t, taps, N, z, a)
*
*             computes the the LFSR sequence from z and places it in a 
*
*   Author  : Azhar Desai
*   Date    : September, 2011
*
****************************************************************************/

/***************************************************************************
*                             INCLUDED FILES
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "algorithm_b.h"
#include "bitarray.h"
#include "pb_probability.h"
#include "pc_paritycheck.h"

/***************************************************************************
*                                FUNCTIONS
***************************************************************************/

/***************************************************************************
*   Function   : algorithm_b
*   Description: Sovle correlation problem using algorithm B
*   Parameters : prob - a prioir correlation probability
*                t - number of taps
*                taps - array of taps
*                N - number of bits in observed sequence
*                z - observed sequence
*                a - LFSR sequence (can optionally give answer to print out more stats)
*                    set it to NULL otherwise
*   Returned   : void 
***************************************************************************/
void algorithm_b(double prob, 
		 unsigned int l,  
		 unsigned int t, 
		 unsigned int taps [], 
		 unsigned int N, 
		 bit_array_t *z,
		 bit_array_t *a)
{
  /* Sanity check
   * 
   * That number of taps is less than lenghth
   * Each tap is less than length
   *
   */
  char logfile [] = "log.tex";
  unsigned int alpha = 5;
  unsigned int maxrounds = 1000;
  unsigned int N_w = 0;
  unsigned int m = pb_calc_m(N, l, t);
  double s_init = pb_calc_s(prob, t);  
  unsigned int h_max = calc_h_max(prob, s_init, m);
  pc_paritychecks_t *parchecks = pc_create(N, t, taps);

  /*compute p_threshold */
  double p_thres = 0.5 * (pb_update(prob, s_init, m, h_max) 
			  + pb_update(prob, s_init, m, h_max + 1));
  
  /* N_thres = U(p, m, h_max) * N */
  unsigned int N_thres = pb_calc_IU(prob, s_init, m, h_max, U_OPT) * N;

  /* Allocate memory */
  double *p = (double *) calloc(N, sizeof(double));
  double *p_star = (double *) calloc(N, sizeof(double));
  double **s = (double **) calloc(N, sizeof(double *));
  
  if (p == NULL || p_star == NULL || s == NULL)
    {
      fprintf(stderr, "Error: Unable to allocate memory for algorithm B to run");
      exit(EXIT_FAILURE);
    }

  /* Initialise parity check array */
  for (unsigned int i = 0; i < N; i++)
    {
      s[i] = (double *) calloc(PC_NUM_PRODS, sizeof(double));

      if (s[i] == NULL)
	{
	  fprintf(stderr, "Error: Unable to allocate memory for S arrays.");
	  exit(EXIT_FAILURE);
	}
    }
  
  /* Print out algorithm details */
  double I_max = pb_calc_IU(prob, s_init, m, h_max, I_OPT);
  double d = (double) N/ (double) l;
  
  printf("------------------------------------------\n");
  printf("Derived Parameters:\n");
  printf("------------------------------------------\n");
  printf("Correlation Probability: \t\t%.3f\n", prob);
  printf("Expected num of corrected bits [N_c]: \t%.5f\n", I_max*N );
  printf("Correction factor [F(p,t,d)]: \t\t%.5f\n", I_max*d);
  printf("Probability threshold: \t\t\t%.3f\n", p_thres);
  printf("Expected num bits below threshold: \t%u\n\n", N_thres);

  printf("------------------------------------------\n");
  printf("Round\tIteration\tN_w \n");
  printf("------------------------------------------\n");

  /* If verbose logging specified print out details to file*/
  FILE *logfp = NULL;
  if (a != NULL)
    {
      if ((logfp = fopen(logfile, "w")) == NULL)
	{
	  fprintf(stderr, "Unable to open logfile: %s", logfile);
	}
      fprintf(logfp, "\\begin{tabular}{llr}\n\\toprule\nRound & Iteration & $\\seq{z}$ \\\\ \n");
    }
  

  /* Algorithm */
  for (unsigned int r = 1; r <= maxrounds; r++)
    {      
      if (logfp != NULL)
	fprintf(logfp, "\\midrule\n%u", r);
      
      /* Reset all probabilities to original constant*/
      for (unsigned int i = 0; i < N; i++)
	{
	  p[i] =  prob;
	}

      /* log start of line */
      printlogline(logfp, 0, N, 0, p, z, a);

      for (unsigned int iter = 1; iter <= alpha; iter++)
	{	  
	  
	  /* Count the number of parity checks satisfied */
	  unsigned int h[N];
	  pc_h(parchecks, z, h);

	  /* For each z compute the new p* probability */
	  pc_parcheck_s(parchecks, z, p, s);	  
	  for (unsigned int i = 0; i < N; i++)
	    {
	      p[i] = pb_update_var_p(p[i], s[i]);
	    }
	  
	  /* Count the number below threshold */
	  N_w = 0;
	  for (unsigned int i = 0; i < N; i++)
	    {
	      if (dbl_gt(p_thres, p[i]))
		{
		  N_w++;
		}	      
	    }
	  
	  printf("%u \t%u \t\t%u\n", r, iter, N_w);
	  printlogline(logfp, iter, N, p_thres, p, z, a);

	  /* If there enough bits below threshold amount then we can 
	     stop iterating, flip them and move onto next round*/
	  if (N_w >= N_thres)
	    {
	      break;
	    }
	}
      
      /* Flip bits below threshold */
      for (unsigned int i = 0; i < N; i++)
	{
	  if (dbl_gt(p_thres, p[i]))
	    {
	      if (BitArrayTestBit(z, i))
		{
		  BitArrayClearBit(z, i);
		}
	      else
		{
		  BitArraySetBit(z, i);
		}
	    }
	}
      
      printf("------------------------------------------\n");
      /* If we haven't flipped any bits this round
	 then the computation will be exactly the same
	 in the next round. Hopefully the program
	 will have converged on a solution, if not
	 the it fails. Otherwise it has converged on
	 an LFSR sequence */
      
      if (N_w == 0)
	{
	  printf("FAILED: No bits have been flipped on the last round of this program, so algorithm cannot proceed.\n");
	  if (pc_check_lfsr(parchecks, z) == 0)
	    {
	      printf("However converged on a solution.\n");
	    }
	  else
	    {
	      printf("Failed to converge on on a solution.\n");
	    }
	  break;
	}
	
      /*check if after successfully flipping bits we have converge on an LFSR sequence */
      if (pc_check_lfsr(parchecks, z) == 0)
      	{
      	  printf("Converged on a LFSR Sequence\n");
      	  break;
      	}
    }
  
   /*Print end of algorithm state and close log file */
  if (logfp != NULL)
    {
      fprintf(logfp, "\\midrule\n");
      printlogline(logfp, 0 , N, -1, p, z, a);
      fprintf(logfp, "\\bottomrule\n\\end{tabular}\n");
      fclose(logfp);
    }
  
    /* 
       free aloocated memory
     */
    pc_destroy(parchecks);
    free(p);
    free(p_star);
    for (unsigned int i = 0; i < N; i++)
      {
    	free(s[i]);
      }
    free(s);
    
}

/***************************************************************************
*   Function   : h_max
*   Description: Compute the maximum h that gives the greatest increase
*                in correct number of bits when flipping them using h as 
*                a threshold
*   Parameters : p - correlastion probability
*                s - probability that y_i = b_i (see report)
*                m - the average number of relations per digit
*   Returned   : 0 <= h_max <= m 
***************************************************************************/
unsigned int calc_h_max(double p,
			double s,
			unsigned int m)
{
  unsigned int h_max = 0;
  double I_max = 0;
  for (unsigned int i = 0; i <= m; i ++)
    {
      double newI = pb_calc_IU(p, s, m, i, I_OPT);
      if (dbl_gt(newI,I_max))
	{
	  h_max = i;
	  I_max = newI;
	}
    }
  return h_max;
}

/***************************************************************************
*   Function   : printlogline
*   Description: Print line of table for current iteration
*   Parameters : logfp - logfile pointer
*                i - iteration number
*                z - current sequence
*                a - answer lfsr sequence
*   Returned   : void
***************************************************************************/
void printlogline(FILE* logfp, 
		  unsigned int i,
		  unsigned int N,
		  double p_thres,
		  double *p,
		  bit_array_t *z, 
		  bit_array_t *a)
{
  if (logfp == NULL)
    return;
  fprintf(logfp, "&%u& \\texttt{", i);
  
  for (unsigned int i = 0; i < N; i++)
    {
      unsigned char bz = BitArrayTestBit(z, i) ? 1 : 0;
      unsigned char ba = BitArrayTestBit(a, i) ? 1 : 0;
      unsigned char c [] = " "; 
      c[0] = (bz == ba) ? '-' : 'x';
   
      if (dbl_gt(p_thres, p[i]))
	{
	  fprintf(logfp, "\\hl{%s} ", c);
	}
      else
	{
	  fprintf(logfp, "%s ", c);
	}
    }
  
  fprintf(logfp, "}\\\\\n");
}


