/***************************************************************************
*                         Computing Parity Checks
*
*   File    : main.c
*   Purpose : Provides main method, which handles IO, and converting
*             data into appropriate form for algorithm to 
*            
*              The filename provided as an argument to main
*              is opened for reading. The parameter values
*              are all seperated by newlines (UNIX) and read 
*              in as in the example, where the comments # indicate
*              which parameter it is.            
*
*              Example input file (everything after # is not included):
*              p           # correlation probability
*              N           # number of observed keystrem bits
*              T           # number of taps of correlated LFSR
*              t_1         # tap 1
*              t_2         # ....
*              ...
*              t_T         # tap T
*              l           # length of correlated answer
*
*             
*   Author  : Azhar Desai
*   Date    : September, 2011
*
****************************************************************************/

/***************************************************************************
*                             INCLUDED FILES
***************************************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "main.h"
#include "algorithm_b.h"

#define PRINT_EXIT(errormsg)  {fprintf(stderr, errormsg); exit(EXIT_FAILURE);}

/***************************************************************************
*                                FUNCTIONS
***************************************************************************/
extern int optind;

/***************************************************************************
*   Function   : main
*   Description: read in parameters of correlation problem from text file
*   Parameters : first command line argument - file describing parameters
*                second command line argument - optional file with answer 
*                written on one line in it. (This is for printing output
*                about the success of the algorithm as it runs.)
***************************************************************************/
int main (int argc, char *argv[])
{  
  unsigned int N = 0, numtaps = 0, length = 0;
  unsigned int *taps = NULL;
  bit_array_t *z, *a = NULL;
  double p= 0;
  char verbose = 0;
  int opt;

  /* Parse options */
  while ((opt = getopt(argc, argv, "l")) != -1)
    {
      switch (opt){
      case 'l': 
	verbose = 1;
	break;

      default:
	m_printusage(argv[0]);
	return EXIT_FAILURE;
      }      
    }
	 
  if (argc - optind < 1 || argc - optind > 2)
    {
      m_printusage(argv[0]);
      return EXIT_FAILURE;
    }

  /* Read in input files.
   * if answer given read that in too.
   */
  m_readinputfile(argv[optind], &p, &N, &length, &numtaps, &taps, &z);
  if (optind + 1 < argc)
    m_readansfile(argv[optind + 1], &a, N);
  
  /* If verbose flag specified -- print out answer */
  if (verbose)
    algorithm_b(p, length, numtaps, taps, N, z, a);
  else
    algorithm_b(p, length, numtaps, taps, N, z, NULL);

  /* if answer file given test if correct, 
     else print sequence out
  */
  if (a != NULL)
    {
      if (BitArrayCompare(z,a) == 0)
	{
	  printf("Sequence is correct.\n");
	}
      else
	{
	  printf("Seqeunce is wrong.\n");
	}
    }
  else
    {
      m_print_btarray(z, N);
    }
  
  free(taps);
  return EXIT_SUCCESS;
}

/***************************************************************************
*   Function   : m_readinputfile
*   Description: open and read in input file into paramters passed into function
*                exits on fail
*   Parameters : p_p - pointer to prob array
*                N_p - pointer to number of keystream bits
*                l_p - pointer to lfsr length
*                numtaps_p, taps_p - pointer number of taps, and taps array
*                z_p - pointer to bit array for storing keystream
***************************************************************************/
void m_readinputfile(const char filename [], 
		     double *p_p,
		     unsigned int *N_p,
		     unsigned int *l_p,
		     unsigned int *numtaps_p,
		     unsigned int **taps_p,
		     bit_array_t **z_p)
{
  FILE *infp;
  infp = m_open_exit_fail(filename," Failed to open input file.\n");

  if (fscanf(infp, "%lf\n%u\n%u\n%u\n", p_p, N_p, l_p,  numtaps_p) < 3)
    {
      PRINT_EXIT("Error: inputfile does not start with:\nN\np\nnumtaps\n");
    }
 
  *taps_p = (unsigned int *) calloc(*numtaps_p, sizeof(unsigned int));
  if (*taps_p == NULL)
    {
      PRINT_EXIT("Failed to allocate memory for tap array in main.c");
    }
 
  unsigned int *taps = *taps_p;
  for (unsigned int i = 0; i < *numtaps_p; i++)
    {
      fscanf(infp, "%u\n", &taps[i]);
    }
  
  
  /*Read in bitstream */
  *z_p = BitArrayCreate(*N_p); 
  bit_array_t *z = *z_p;
  BitArrayClearAll(z);
  
  for (unsigned int i = 0; i < *N_p; i++)
    {
      int c = fgetc(infp);
      switch (c)
	{
	case '0':
	  BitArrayClearBit(z, i);
	  break;
	case '1':
	  BitArraySetBit(z, i);
	  break;
	default:
	  PRINT_EXIT("Failed to read in bit array stream.");
	}
    }

  fclose(infp);
}

/***************************************************************************
*   Function   : m_readansfile
*   Description: open and read in one line asnwer file 
*   Parameters : filename - filename
*                a_p - pointer to lfsr sequence bit array
*                N - number of bits to read in
***************************************************************************/
void m_readansfile(const char filename [], bit_array_t **a_p, unsigned int N)
{
    FILE *ansfp;
    ansfp = m_open_exit_fail(filename," Failed to open input file.\n");
    /*Read in bitstream */
    *a_p = BitArrayCreate(N); 
    bit_array_t *a = *a_p;
    BitArrayClearAll(a);
  
    for (unsigned int i = 0; i < N; i++)
      {
	int c = fgetc(ansfp);
	switch (c)
	  {
	  case '0':
	    BitArrayClearBit(a, i);
	    break;
	  case '1':
	    BitArraySetBit(a, i);
	    break;
	  default:
	    PRINT_EXIT("Failed to read in bit array stream.");
	  }
      }
    fclose(ansfp);
}

/***************************************************************************
*   Function   : m_open_exit_fail
*   Description: open file at filename provided; exit if failed.
*   Parameters : filenae - to open
*                errormsg - message to print if error occurs
*   Returns    : pointer to successfully opened file
***************************************************************************/
FILE * m_open_exit_fail(const char filename [], const char errormsg [])
{
  FILE *fp = fopen(filename, "r");
  if (fp == NULL)
    {
      PRINT_EXIT(errormsg);
    }
  return fp;
}

/***************************************************************************
*   Function   : m_printusage
*   Description: print usage of progame
*   Parameters : progname - programe name, usually argv[0]
***************************************************************************/
void m_printusage(const char progname[])
{
  fprintf(stderr, "Usage: %s [-l] inputfile [answerfile]\n", progname);
  /* Print stuff about file format here */
}

/***************************************************************************
*   Function   : m_print_btarray
*   Description: print bit array to stdout
*   Parameters : z - bitarray
*                N - length of z
***************************************************************************/
void m_print_btarray(bit_array_t *z, unsigned int N)
{
  for(unsigned int i = 0; i < N; i++)
    {
      if (BitArrayTestBit(z, i))
	putchar('1');
      else
	putchar('0');
    }
  putchar('\n');
}


