/***************************************************************************
*                         Probability Module
*
*   File    : pb_probability.c
*   Purpose : Provides functions to computes probabilities needed to solve
*             the correlation problem
*             
*   Author  : Azhar Desai
*   Date    : September, 2011
*
****************************************************************************/

/***************************************************************************
*                             INCLUDED FILES
***************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include<gsl/gsl_sf_gamma.h>

#include "pb_probability.h"
#include "pc_paritycheck.h"

/***************************************************************************
*                                FUNCTIONS
***************************************************************************/

/***************************************************************************
*   Function   : pb_calc_m
*   Description: Compute the average number of parity check equations
*                available per a bit
*   Parameters : N - number of observed bits
*                k - correlated LFSR length
*                t - number of taps of correlated LFSR
*   Returned   : m
***************************************************************************/
unsigned int pb_calc_m (unsigned int N, 
			unsigned int k, 
			unsigned int t)
{
  /*
    Return m = 0 on invalid data 
     See correponding test data for more details
     ../test/check_pb_probability.c
  */
  if (N == 0 || k == 0 || t == 0 || t > k)
    return 0;

  double temp = (double)N/(double)(2*k);
  if (temp < 1)
    return 0;

  /* Getting log2 = log / log(2) with ansi c*/
  temp = log(temp)/log(2)*(t+1);
  return (unsigned int)(temp);
}


/***************************************************************************
*   Function   : pb_calc_s
*   Description: Compute s the probability that b_i = y_i (see report)
*                s(p, t) = ps(p, t-1) + (1-p)(1-s(p, t-1))
*                s(p, 1) = p
*   Parameters : p - correlation probability
*                t - number of taps of correlated LFSR
*   Returned   : s
***************************************************************************/
double pb_calc_s (double p, unsigned int t)
{
  /* Assert sane probablility value, and at least one tap */
  ASSERT_PROB(p);
  assert(t >= 1);
  
  double curr_s = p;
  unsigned int i;
  for (i = 1; i < t; i++)
    curr_s = p * curr_s + (1-p)*(1-curr_s);
  return curr_s;
}

/***************************************************************************
*   Function   : pb_calc_s_var_p
*   Description: Compute s the probability that b_i = y_i (see report)
*                when p is variable
*   Parameters : p [] - array of length N correlation probabilities for each bit
*                taps [] - array of length t + 1 representing current parity 
*                check equation.
*                polyterms - number of terms in current parity check equation
*                excl_t - the current tap being examined, ie taps[excl_t] is
*                ignored.
*   Returned   : s
***************************************************************************/
double pb_calc_s_var_p(double *p,
		      unsigned int *taps,
		      unsigned int polyterms,
		      unsigned int excl_t)
{
  double s;
  unsigned int starti;
  
  if (excl_t == 0)
    {
      s = p[taps[1]];
      starti = 2;
    }
  else
    {   
      s = p[taps[0]];
      starti = 1;
    }
        
  for (unsigned int i = starti; i < polyterms; i++)
    {
      if (i != excl_t)
	{
	  double nextp = p[taps[i]];
	  s = nextp*s + (1-nextp)*(1-s);
	}
    }
  
  return s;
}


/***************************************************************************
*   Function   : pb_ps
*   Description: Compute a commonly used probabilty expression
*
*                Z_TRUE: p*s^i*(1-s)^(m-i)
*                Z_FALSE: (1-p)*(1-s)^i*s^(m-i)
*                Z_SUM: Z_TRUE + ZFALSE
*
*                The interpretaion of this is the probability 
*                that the parity check for  z is satifised when:
*                
*                 Z_TRUE: z == a and exactly i of m of the other terms, 
*                         L_i are 0
*                 Z_FALSE: z != a and exactly m-i parity checks, 
*                         L_i are 0
*                 Z_SUM: either of Z_TRUE or Z_FALSE holds
*
*   Parameters : p - correlation probability
*                s - probability that y_i = b_i (see report)
*                m - average number of parity checks available per a bit
*                i - number of relations satisfied
*                opt - whether Z_{TRUE,FALSE,SUM}
*   Returned   : probability value
***************************************************************************/
double pb_ps (double p, 
	      double s, 
	      unsigned int m, 
	      unsigned int i, 
	      PSopt_t opt)
  {
  /* Assert p, and s, probalilties  sanity checks on values */
  ASSERT_PROB(p);
  ASSERT_PROB(s);
  assert(i <= m); /* satisfy less than total number of checks to satisfy*/
  assert(opt == Z_TRUE || opt == Z_FALSE || opt == Z_SUM); /* for the kids */
  
  /* 
   * In the case where s is 0 or 1 the probabliltiy 
   * differs from the expression 
   */
  
  double ans = 0;
  if (opt == Z_TRUE || opt == Z_SUM)
    {
      if (dbl_equals(s,0))
	{
	  ans += (i == 0) ? p : 0;
	}
      else if (dbl_equals(s,1))
	{
	  ans += (i == m) ? p : 0;
	}
      else
	{
	  ans += p * pow(s, i) * pow(1-s, m-i);
	}
    }
  
  if (opt == Z_FALSE || opt == Z_SUM)
    {

      if (dbl_equals(s,0))
	{
	  ans += (i == m) ? (1 - p) : 0;
	}
      else if (dbl_equals(s,1))
	{
	  ans += (i == 0) ? (1 - p) : 0;
	}
      else
	{
	  ans += (1-p) * pow( 1 - s, i) * pow(s, m-i);
	}
    }
  
  return ans;
}

/***************************************************************************
*   Function   : udpate_prob
*   Description: Update probability that z = a, given the new information
*                that the bit satifies h of m parity check equations
*   Parameters : p - prior correlation probability
*                s - probability that b_i = y_i (see report)
*                m - average number of relations per bit
*                h - number parity check equations bit satisfied
*   Returned   : Updated probability
***************************************************************************/
double pb_update(double p, 
		 double s,
		 unsigned int m,
		 unsigned int h)
{
  double ans = 0;
  if (!(dbl_equals(p, 0) || 
	dbl_equals(s, 0)  || 
	dbl_equals(s, 1) ))
    {
      ans = pb_ps(p, s, m, h, Z_TRUE) / pb_ps(p, s, m, h, Z_SUM);
    }
  return ans;
}

/***************************************************************************
*   Function   : udpate_prob
*   Description: Update probability that z = a, given the new information
*                that the bit satifies h of m parity check equations
*   Parameters : p - prior correlation probability of z = a
*                s_vals  - array of length 4 which encodes
*                the (generalised) probability that b_i = y_i (see report)
*   Returned   : Updated probability
***************************************************************************/
double pb_update_var_p(double p,
		       double *svals)
{
  assert(svals != NULL);
  double t = p * svals[PC_SATISFIED_PROD] * svals[PC_UNSATISFIED_PROD_COMPLEMENT];
  t = t / (t + (1-p)*svals[PC_SATISFIED_PROD_COMPLEMENT] * svals[PC_UNSATISFIED_PROD]);
  return t;
}

/***************************************************************************
*   Function   : I
*   Description: Compute relative increase ( -1 <= I <= 1) of the correct 
*                number of bits after flipping for various thresholds h
*   Parameters : p - correlation probability
*                s - probability that y_i = b_i (see report)
*                m - the average number of relations per digit
*                h - digits satisying < h relaitons are complemented
*   Returned   : double I 
*                (can be negative, if more incorrect bits after flipping)
***************************************************************************/
double pb_calc_IU(double p, 
		  double s, 
		  unsigned int m, 
		  unsigned int h,
		  IUopt_t opt)
{
  /* Assert basic sanity checks on values */
  ASSERT_PROB(p);
  ASSERT_PROB(s);
  assert(h <= m); /* Can't satisfy more than m relations */
  assert((opt == I_OPT) || (opt == U_OPT)); /* for the kids */

  double ans = 0;
  
  unsigned int i;
  for (i = 0; i <= h; i++)
    {
      /* Add summand times m choose k */
      double fp1 = pb_ps(p, s, m, i, Z_FALSE);
      double fp2 = pb_ps(p, s, m, i, Z_TRUE);
      double fp = (opt == I_OPT) ? fp1 - fp2 : fp1 + fp2;
      ans += gsl_sf_choose(m, i)*fp;

    }
  return ans;
}
