/***************************************************************************
*                         Computing Parity Checks
*
*   File    : pc_paritycheck.c
*   Purpose : Provides functions for generating parity checks and testing
*             how many checks each bit of keystream satisfies
*
*           Example: if we have polynomial 1 + X + X^2 and the keystream 
*                    length, N, is 5, then the polynomial 
*                    1 + X^2 + X^4 is also generated.
*                    
*                    Polynomials are represented as arrays of unsigned int as 
*                    follows:
*                     
*                    1 + X   + X^2 ==> a_0 + a_1 + a_3 = 0 
*                    1 + X^2 + X^4 ==> a_0 + a_2 + a_4 = 0
*
*                    The arrays store the initial positions of the equations
*                    as:
*                           [0, 1, 3] and [0, 2 , 4]
*                    and then increment the positions of the terms to get
*                    the next equation 
*
*
*   Author  : Azhar Desai
*   Date    : September, 2011
*
****************************************************************************/


/***************************************************************************
*                             INCLUDED FILES
***************************************************************************/
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <errno.h>

#include "bitarray.h"
#include "pc_paritycheck.h"
#include "pb_probability.h"

/***************************************************************************
*                            TYPE DEFINITIONS
***************************************************************************/
struct pc_paritychecks_t
{
  unsigned int N;
  unsigned int t;
  unsigned int num_polys;
  /* Ordered from highest to lowest degree
     so that poly[i][0] is the degree of the i^th polynomial
   */
  unsigned int  **polys;
  unsigned int *degree_polys;
};

/***************************************************************************
*                                FUNCTIONS
***************************************************************************/

/***************************************************************************
*   Function   : pc_create
*   Description: Create structure to store and generate all appropriate
*                feedback  polynomials.
*   Parameters : N - length of observed keystream
*                t - number of taps (= number of terms in polynomials)
*                taps - tap positions of LFSR feedback polynomial
*   Returned   : pointer to allocated new srtuct NULL if struct not
*                allocated.
***************************************************************************/
pc_paritychecks_t *pc_create(unsigned int N, 
			     unsigned int t, 
			     unsigned int taps [])
{
  
  /* 
   * Count the number of feedbackup polynomials:
   * 1. find higest degreee (tap) in given  polynomial.
   * 2. count how many times we can  square it without exceeding N
   */
  unsigned int maxtap = 0;
  unsigned int num_polys = 0;
  unsigned int origpolydegree = 0;

  for (unsigned int i = 0; i < t; i++)
    if (maxtap < taps[i])
      maxtap = taps[i];

  origpolydegree = maxtap;
      
  while (maxtap < N)
    {
      num_polys++;
      /* Square maxtap */
      maxtap <<= 1; 
    }

  
  /* Allocate memory for:
   *
   * 1. struct
   * 2. arrays degrees of polynomials, and bitarrays
   * 3. each polynomial array
   */  

  /* 1. struct */
  pc_paritychecks_t *pc = NULL;
  pc = (pc_paritychecks_t *)malloc(sizeof(pc_paritychecks_t));

  if (pc == NULL)
    {
      fprintf(stderr, "Error: Unable to allocate memory for pc_paritycheck_t struct");
      exit(EXIT_FAILURE);
    }

  /* 2. Arrays */
  pc->degree_polys = NULL;
  pc->polys = NULL;

  pc->degree_polys = (unsigned int*) calloc(num_polys, sizeof(unsigned int *));
  pc->polys = (unsigned int **) calloc(num_polys ,sizeof(unsigned int *));
  
  if ((pc->degree_polys == NULL) || (pc->polys == NULL))
    {
      fprintf(stderr, "Error: Unable to allocate memory inside pc_paritycheck_t struct.");
      pc_destroy(pc);
      return NULL;
    }
  
  /*3. Each Polynomial Array */
  for (unsigned int i = 0; i < num_polys; i++)
    {
      pc->polys[i] = (unsigned int *) calloc(t + 1, sizeof(unsigned int));
      if (pc->polys[i] == NULL)
	{
	  fprintf(stderr, "Error: Failed to alloc mem in for polynomial in when creating pc_parity_check_t ");
	  pc_destroy(pc);
	  return NULL;
	}
    }

  /* 
   * Initialise values of struct 
   * 1. Copy constants
   * 2. Compute degrees of each polynomial
   * 3. Compute the powers of the terms of each polynomial
   */
  
  pc->N = N;
  pc->t = t;
  pc->num_polys = num_polys;
  
  pc->degree_polys[0] = origpolydegree;
  for (unsigned int i = 1; i < num_polys; i++)
    {
      /* Square previous degree */
      pc->degree_polys[i] = pc->degree_polys[i-1] << 1;
    }
  
  /* Note: we're assuming here that the LFSR is maximal
     ie the max degree of feedback polynomial is equal 
     so there is no need to shift a sequence
     10100000 => to 0010100 in the case where the LFSR length
     is 5, but degree of feedback polyonmial is 3.
  */
  
  /* Compute the powers of the terms in the polynomial */
  for (unsigned int i = 0; i < pc->num_polys; i++)
    {
      /* Set all bits to 0 */
      
      /* Set the bit corresponding to the constant 1 in the polynomial*/
      /* BitArraySetBit(pc->polys[i], pc->degree_polys[i]); */
      
      /* First term is highest degree term */
      pc->polys[i][0] = pc->degree_polys[i];
      
      for (unsigned int j = 0; j < pc->t; j++)
	{
	  /* BitArraySetBit(pc->polys[i], pc->degree_polys[i] - taps[j]); */
	  pc->polys[i][j+1] = pc->degree_polys[i] - taps[j];
	  /* Square polynomial term for next polynomial */
	  taps[j] <<= 1;
	} 
    }
  return pc;
}


/***************************************************************************
*   Function   : pc_destroy
*   Description: Destory allocated parity checks 
*   Parameters : pchecks - struct to allocated parity check
***************************************************************************/
void pc_destroy(pc_paritychecks_t *pchecks)
{
  /* destroy all bit arrays */
  if (pchecks != NULL)
    {
      if (pchecks->polys != NULL)
	{
	  for (unsigned int i = 0; i < pchecks->num_polys; i++)
	    {
	      if (pchecks->polys[i] != NULL)
		{
		  free(pchecks->polys[i]);
		}
	    }
      
	  free(pchecks->polys);
	}
      free(pchecks);
    }
}


/***************************************************************************
*   Function   : pc_numpolys
*   Description: access number of polynomials contained in struct
*   Parameters : pchecks - struct to get number of polys
***************************************************************************/
unsigned int pc_numpolys(pc_paritychecks_t *pchecks)
{
  assert(pchecks != NULL);
  return pchecks->num_polys;
}

/***************************************************************************
*   Function   : pc_N
*   Description: Access number of keystream bits, N, belonging to given struct
*   Parameters : pchecks - struct from which to get N.
***************************************************************************/
unsigned int pc_N(pc_paritychecks_t *pchecks)
{
  assert(pchecks != NULL);
  return pchecks->N;
}

/***************************************************************************
*   Function   : pc_polys
*   Description: retrive polynomials from struct
*   Parameters : pointer to pc_paritychecks_t struct
*   Returned   : pointer to array representing polynomials
***************************************************************************/
unsigned int **pc_polys(pc_paritychecks_t *pchecks)
{
  /*Sanity Check*/
  assert(pchecks != NULL);
  return pchecks->polys;
}


/***************************************************************************
*   Function   : pc_h
*   Description: Compute the number of relations satisfied for each bit
*                and store in given array
*   Parameters : parchecks - paritycheck polynomials generated
*                z - observed keystream sequence
*                h [] - array to store the number of relations computed for 
*                        each bit
*		Note: Length of z must equal parchecks->N and length of h
***************************************************************************/
void pc_h(pc_paritychecks_t *parchecks,
		   bit_array_t *z,
		   unsigned int h [])
{
  /* For safety maybe assert bit_array_t length - need to add method */
  
  /* Initialise h to 0 */
  for (unsigned int i = 0; i < parchecks->N; i++)
    {
      h[i] = 0;
    }
  
  /* Loop through each shift of each polynomial 
   * Effectively shifting the linear reccurence equation one term up
   */
  for (unsigned int i = 0; i < parchecks->num_polys; i++)
    {
      /* Sanity Check*/
      assert(parchecks->N >= parchecks->degree_polys[i]);
      
      unsigned int *currtaps = (unsigned int*)calloc(parchecks->t + 1, sizeof(unsigned int));
      
      if (currtaps == NULL)
  	{
  	  fprintf(stderr, "Error: Failed to allocate memory when checking parity check equations");
  	  exit(EXIT_FAILURE);
  	}
      
      /* Copy current polynomial taps */
      for (unsigned int j = 0; j < parchecks->t + 1; j++)
	{
	  currtaps[j] = parchecks->polys[i][j];
	}

      
      /* compute the number of shifts of the polynomial we can make*/
      unsigned int shifts = parchecks->N - parchecks->degree_polys[i];
      
      for (unsigned int j = 0; j < shifts; j++)
  	{
  	  /*
	   * Test parity check equation,if is satisfied increment 
	   * the count of all bits involved in it.
	   * Lastly, shift equation one bit up the sequence
	   */
	  
  	  unsigned char parity = 0;
  	  
	  for (unsigned int j = 0; j < parchecks->t + 1; j++)
	    {
	      if (BitArrayTestBit(z, currtaps[j]))
		{
		  parity += 1; 
  	 	  parity = (parity == 2) ? 0 : parity;
		}
	    }
	  
  	  if (!parity)
  	    {
  	      for (unsigned int i = 0; i < parchecks->t + 1; i++)
  		{
  		  h[currtaps[i]] += 1;
  		}
  	    }
	  
  	  for (unsigned int i = 0; i < parchecks->t + 1; i++)
  	    {
  	      currtaps[i]++;
  	    }
  	}
    }
}

/***************************************************************************
*   Function   : pc_parcheck_s
*   Description: Compute the number of relations satisfied and use this
*                to compute the s_i values needed to update probabilities
*   Parameters : parchecks - paritycheck polynomials generated
*                z - observed keystream sequence
*                p - array where p[i] is prob(z_i = a_i)
*                s -  a 2D array where each s[i] is and array of four
*                    doubles to store answer
*
*                s[i][PC_SATISFIED_PROD] = s_0 * s_1 * ... * s_h
*                s[i][PC_SATISFIED_PROD_COMPLEMENT] = (1-s_0)* ... * (1-s_h)
*                
*                where bit i satisfies its parity check equations 0 .. h
*                The other two members of s[i] are defined analogously
*                for the when i does not satisfy it's parity check equation
*              
*                See report for more details.
*                 
*		Note: Length of z must equal parchecks->N and length of s
***************************************************************************/
void pc_parcheck_s(pc_paritychecks_t *parchecks,
		   bit_array_t *z,
		   double p [],
		   double **s)
{
  
  unsigned int polyterms = parchecks->t + 1;
  unsigned int *currtaps = (unsigned int*)calloc(polyterms, 
						 sizeof(unsigned int));
  if (currtaps == NULL)
    {
      fprintf(stderr, "Error: Failed to allocate memory when checking parity check equations");
      exit(EXIT_FAILURE);
    }

  /* Initialise all values of s[][] to 1*/
  for (unsigned int i = 0; i < parchecks->N; i++)
    {
      s[i][PC_SATISFIED_PROD] = 1;
      s[i][PC_SATISFIED_PROD_COMPLEMENT] = 1;
      s[i][PC_UNSATISFIED_PROD] = 1;
      s[i][PC_UNSATISFIED_PROD_COMPLEMENT] = 1;
    }
  
   for (unsigned int i = 0; i < parchecks->num_polys; i++)
    {
      /* Sanity Check*/
      assert(parchecks->N >= parchecks->degree_polys[i]);
            
      /* Copy current polynomial taps - so that it is not modified*/
      for (unsigned int j = 0; j < polyterms; j++)
	{
	  currtaps[j] = parchecks->polys[i][j];
	}

      /* compute the number of shifts of the polynomial we can make*/
      unsigned int shifts = parchecks->N - parchecks->degree_polys[i]; 
      for (unsigned int j = 0; j < shifts; j++)
  	{
  	  /*
	   * Test parity check equation,if is satisfied increment 
	   * the count of all bits involved in it.
	   * Lastly, shift equation one bit up the sequence
	   */
	  
  	  unsigned char parity = 0;
	  for (unsigned int k = 0; k < polyterms; k++)
	    {
	      if (BitArrayTestBit(z, currtaps[k]))
		{
		  parity += 1; 
  	 	  parity = (parity == 2) ? 0 : parity;
		}
	    }
	  
	  for (unsigned int k = 0; k < polyterms; k++)
	    {      
	      double curr_s = pb_calc_s_var_p(p, currtaps, polyterms, k); 

	      unsigned int bit = currtaps[k];
	      /*satisfied parity check */
	      if (!parity)
		{
		  s[bit][PC_SATISFIED_PROD] *= curr_s;
		  s[bit][PC_SATISFIED_PROD_COMPLEMENT] *= (1 - curr_s);
		}
		/* NOT satisfied parity check */
	      else
		{
		  s[bit][PC_UNSATISFIED_PROD] *= curr_s;
		  s[bit][PC_UNSATISFIED_PROD_COMPLEMENT] *= (1 - curr_s);
		}
	    }
  	  
  	  for (unsigned int k = 0; k < polyterms; k++)
  	    {
  	      currtaps[k] += 1;
  	    }
  	}
    }
   
   free(currtaps);
}


/***************************************************************************
*   Function   : pc_check_lfsr
*   Description: Compute the number of relations satisfied and use this
*                to compute the s_i values needed to update probabilities
*   Parameters : parchecks - struct describing LFSR
*                z - sequence to test whether it is an LFSR
*   Returned   : 0 if successful, 1 if not
***************************************************************************/
unsigned char pc_check_lfsr(pc_paritychecks_t *parchecks, bit_array_t *z)
{
  /* Use first polynomial and check this */
  unsigned char ret = 3;
  unsigned int polyterms = parchecks->t + 1;
  unsigned int shifts = parchecks->N - parchecks->degree_polys[0];
  unsigned int *poly = calloc(polyterms, sizeof(unsigned int));
  
  if (poly == NULL)
    {
      fprintf(stderr, "Error: Unable to allocate memory to check if lfsr");
      exit(EXIT_FAILURE);

    }
  
  /* Copy the first polynomial */
  for (unsigned int i = 0; i < polyterms; i++)
    {
      poly[i] = parchecks->polys[0][i];
    }
  
  /* Check polynomial then shift up */
  for (unsigned int i = 0; i < shifts; i++)
    {
           
      unsigned char parity = 0;
      for (unsigned int k = 0; k < polyterms; k++)
	{
	  if (BitArrayTestBit(z, poly[k]))
	    {
	      parity += 1; 
	      parity = (parity == 2) ? 0 : parity;
	    }
	}
      
      /* if check fails, then is not lfsr seq -> return*/
      assert( parity == 0 || parity == 1);
      if (parity == 1)
	{
	  return 1;
	}
      /* shift up */
      for (unsigned int k = 0; k < polyterms; k++)
	{
	  poly[k] += 1;
	}
    }

  /* Set to a non-zero value earlier to 
     ensure that tests fail appropriately */
  ret = 0;
  return ret;
}
