#include <stdlib.h>
#include <check.h>
#include "../src/pb_probability.h"

/*
 * Checks pb_calc_m
 *
 * Average number of relations per bit
 */
START_TEST (test_pb_calc_m)
{
  fail_unless(pb_calc_m(200,20,18) == 44, 
	      "Relatively normal values" );
  fail_unless(pb_calc_m(0,0,0) == 0,
	      "Get 0 equations if any of N, k, or t = 0");
  fail_unless(pb_calc_m(4,29,4) == 0, 
	      "Get 0 if N/2k < 1 : haven't observed enough more than twice LFSR length.");
    fail_unless(pb_calc_m(400,20,100) == 0, "Get 0 if number of taps exceeds LFSR length.");
}
END_TEST

/*
 * Checks pb_calc_s (with constant p)
 */
START_TEST(test_pb_calc_s)
{
  fail_unless(dbl_equals(pb_calc_s(0.75, 3), 0.5625),
  	      "Relatively normal values for s.");
  fail_unless(dbl_equals(pb_calc_s(0.23, 5), 0.4770417488),
  	      "Relatively normal values for s. (2)");

}
END_TEST

/* Check pb_calc_s (with variable p)*/
START_TEST(test_pb_calc_s_var_p)
{
  double p [] = {0.67, 0.45, 0.23, 0.78, 0.56, 0.89};
  unsigned int taps [] = {2,3,4};
  unsigned int t = 3;
  unsigned int excl_t = 1;
  
  double computed = pb_calc_s_var_p(p, taps, t, excl_t);
  double ans = p[2]*p[4] + (1-p[2])*(1-p[4]);
  
  fail_unless(dbl_equals(computed, ans),
	      "Expected %.8f but got %8.f",
	      ans, computed);
  
  computed = pb_calc_s_var_p(p, taps, t, 0);
  ans = p[3]*p[4] + (1 - p[3])*(1-p[4]);
  
  fail_unless(dbl_equals(computed, ans),
	      "Expected %.8f but got %8.f",
	      ans, computed);

}
END_TEST

START_TEST(test_pb_calc_s_var_p_1)
{
  double p [] = {0.67, 0.45, 0.23, 0.78, 0.56, 0.89};
  unsigned int taps [] = {2,3,4,1};
  unsigned int t = 4;
  unsigned int excl_t = 1;
  
  double computed = pb_calc_s_var_p(p, taps, t, excl_t);
  double ans = p[2]*p[4] + (1-p[2])*(1-p[4]);
  ans = ans * p[1] + (1-ans)*(1 - p[1]);
  
  fail_unless(dbl_equals(computed, ans),
	      "Expected %.8f but got %8.f",
	      ans, computed);
  
}
END_TEST

/* 
 * Checks ps for the option that Z_TRUE
 * The crucial formula
 *
 */
START_TEST(test_ps_ztrue)
{
  double ans, computed;
  
  /* Trivial cases */
  fail_unless(pb_ps(0, 0.34, 23, 12, Z_TRUE) == 0, 
	      "PS return 0 on Z_Sum if p = 0");
  fail_unless(pb_ps(0.5, 0, 23, 13, Z_TRUE) == 0, 
	      "Return 0 on Z_Sum if s = 0");
  fail_unless(pb_ps(0.7, 1, 13, 4, Z_TRUE) == 0, 
	      "Return 0 on Z_Sum if s = 1");
  
  /* Handle cases where s = 1 or 0 */
  fail_unless(dbl_equals(pb_ps(0.75, 0, 23, 0, Z_TRUE), 0.75), 
	      "PS: Z_TRUE return p if s = 0 and i=0");

  fail_unless(dbl_equals(pb_ps(0.75, 0, 23, 23, Z_TRUE),0), 
	      "PS: Z_TRUE return 0 if s = 0 and i=m");

  fail_unless(dbl_equals(pb_ps(0.75, 1, 23, 0, Z_TRUE),0), 
	      "PS: Z_TRUE return 0 if s = 1 and i=0");

  fail_unless(dbl_equals(pb_ps(0.75, 1, 23, 23, Z_TRUE), 0.75), 
	      "PS: Z_TRUE return p if s = 1 and i=m");

  fail_unless(dbl_equals(pb_ps(0.75, 0, 23, 12, Z_TRUE),0), 
	      "PS: Z_TRUE return 0 if s = 0 and 0 < i < m");

  fail_unless(dbl_equals(pb_ps(0.75, 1, 23, 11, Z_TRUE),0), 
	      "PS: Z_TRUE return 0 if s = 1 and 0 < i < m");


  /* Non trivial cases*/
  ans = 0.00181398528;
  computed = pb_ps(0.3, 0.6, 10, 10, Z_TRUE);
  fail_unless(dbl_equals(ans, computed), 
	      "Relatively normal values when i = m \nAmount was %.7f instead of %.7f", 
	      computed, ans);

  ans = 0.000000017;
  computed = pb_ps(0.7, 0.34, 23, 12, Z_TRUE);
  fail_unless(dbl_equals(ans, computed),
	       "Relatively normal values when i < m \nAmount was %.7f instead of %.7f", computed, ans );

  /* Added in extra case - that appeared to give trouble 
   but it still passees*/
  ans = 0.216;
  computed = pb_ps(1, 0.4, 3, 0, Z_TRUE);
  fail_unless(dbl_equals(ans, computed),
	       "p=1, i=0 test \nAmount was %.7f instead of %.7f", computed, ans );

}
END_TEST

START_TEST(test_ps_zfalse)
{
  double ans, computed;
  
  /* Trivial Cases */
  fail_unless(dbl_equals(pb_ps(1, 0.34, 23, 12, Z_FALSE),0), 
	      "PS return 0 on Z_FALSE if p = 1");

  /* Handle cases where s = 1 or 0 */
  fail_unless(dbl_equals(pb_ps(0.75, 0, 23, 0, Z_FALSE), 0), 
	      "PS: Z_FALSE return 0 if s = 0 and i=0");

  fail_unless(dbl_equals(pb_ps(0.75, 0, 23, 23, Z_FALSE), 1 - 0.75), 
	      "PS: Z_FALSE return (1-p) if s = 0 and i=m");

  fail_unless(dbl_equals(pb_ps(0.75, 1, 23, 0, Z_FALSE), 1 - 0.75), 
	      "PS: Z_FALSE return (1-p) if s = 1 and i=0");

  fail_unless(dbl_equals(pb_ps(0.75, 1, 23, 23, Z_FALSE), 0), 
	      "PS: Z_FALSE return 0 if s = 1 and i=m");

  fail_unless(dbl_equals(pb_ps(0.75, 0, 23, 12, Z_FALSE),0), 
	      "PS: Z_FALSE return 0 if s = 0 and 0 < i < m");

  fail_unless(dbl_equals(pb_ps(0.75, 1, 23, 11, Z_FALSE),0), 
	      "PS: Z_FALSE return 0 if s = 1 and 0 < i < m");

  /* Non-trivial cases */
  ans = 0.000341967;
  computed = pb_ps(0.7, 0.3, 19, 19, Z_FALSE);
  fail_unless(dbl_equals(ans, computed),
	      "For i = m: got %.8f instead of %.8f", computed, ans);

  ans = 0.000078364;
  computed = pb_ps(0.85, 0.6, 14, 1, Z_FALSE);;
  fail_unless(dbl_equals(ans, computed),
	      "For i < m: got %.8f instead of %.8f", computed, ans);  
  
}
END_TEST


START_TEST(test_ps_zsum)
{
  /* Not needed to test as rigourously, because
     the other two options are essentially what 
     make this one work.*/
  
  double ans, computed;
  /*Trivial Cases*/
  fail_unless(pb_ps(0.5, 1, 23, 12, Z_SUM) == 0, 
	      "PS return 0 on Z_Sum if s = 1");
  
  /*Non-trivial check if sum computes correctly */
  ans = 0.000134775;
  computed = pb_ps(0.6, 0.3, 12, 5, Z_SUM);
  fail_unless( dbl_equals(ans, computed), 
	      "PS Z_Sum test on relatively normal vals");
  
}
END_TEST

/* Test updating probability*/
START_TEST(test_pb_udpate0)
{
  double ans = 0.8709677419354;
  double computed = pb_update(0.75, 0.6, 10, 6);
  fail_unless( dbl_equals(ans, computed),
	       "Output %.15f instead of %.15f",
	       computed,
	       ans);
	       
  
}
END_TEST

/* Test updating probability*/
START_TEST(test_pb_udpate1)
{
  /* Some trivial values */
  fail_unless(dbl_equals(pb_update(0, 0.6, 19, 12), 0),
	      "%.15f", pb_update(0, 0.6, 19, 12));
  fail_unless(dbl_equals(pb_update(0.5, 1, 25, 10), 0),
	      "%.15f", pb_update(0.5, 1, 25, 10));
  fail_unless(dbl_equals(pb_update(0.89, 0, 13, 7), 0),
	      "%.15f", pb_update(0.89, 0, 13, 7));

}
END_TEST

/* Test updating probability*/
START_TEST(test_pb_udpate2)
{
  /* Big values */
  double ans = 0.9999711726833;
  double computed = pb_update(0.52, 0.76, 11, 10);
  fail_unless( dbl_equals(ans, computed),
	       "Output %.15f instead of %.15f",
	       computed,
	       ans);
 
}
END_TEST

/* Test Update probability with variable p*/
START_TEST(test_pb_update_var_p)
{
  double svals []  = {0.2, 0.45, 0.23, 0.7};
  double ans = 0.75*svals[0]*svals[3] / (0.75*svals[0]*svals[3] + 0.25*svals[1] * svals[2]);
  double computed = pb_update_var_p(0.75, svals);
  fail_unless( dbl_equals(ans, computed),
	       "Output %.8f instead of %.8f",
	       computed,
	       ans); 
}
END_TEST

/* 
 * Checks pb_calc_IU(p, m, h)
 *
 */
START_TEST (test_I)
{
  double ans;
  double computed;
  
  /* Trivial check -- Is this correct or should it be 0?*/
  ans = 0.25;
  computed = pb_calc_IU(0.75, 1, 5, 3, I_OPT);
  fail_unless(dbl_equals(computed, ans),
	      "Return 0 on s = 1\n returned %.08f instead of %.08f", computed, ans);

  /* Non-trivial checks */
  ans = -0.936;
  computed = pb_calc_IU(1, 0.4, 3, 2, I_OPT);
  fail_unless(dbl_equals(ans, computed),
  	      "Test p = 1\n returned %.8f where %.8f expected", computed, ans);
  
  ans = -0.47;
  computed = pb_calc_IU(0.75, 0.4, 2, 1, I_OPT);
  fail_unless(dbl_equals(ans, computed),
  	      "Test Norm values\n returned %.8f where %.8f expected", computed, ans);

}
END_TEST

/* Test computing U using pb_calc_IU(p, m, h) */
START_TEST (test_U)
{
  double ans = 0.79;
  double computed = pb_calc_IU(0.75, 0.4, 2, 1, U_OPT);
  fail_unless(dbl_equals(ans, computed),
  	      "Test Norm values\n returned %.8f where %.8f expected", computed, ans);
  
}
END_TEST

START_TEST (test_U_oneloop)
{
  double ans = 0.31;
  double computed = pb_calc_IU(0.75, 0.4, 2, 0, U_OPT);
  fail_unless(dbl_equals(ans, computed),
  	      "Test Norm values\n returned %.8f where %.8f expected", computed, ans);
  
}
END_TEST



Suite *
pb_probability_suite(void)
{
  Suite *s = suite_create("Probability Library");
  
  /*Add test_pb_calc_m test cases */
  TCase *tc_pb_calc_m = tcase_create("Calculate m");
  tcase_add_test(tc_pb_calc_m, test_pb_calc_m);
  suite_add_tcase(s, tc_pb_calc_m);

  /*Add test_pb_calc_s test cases*/
  TCase *tc_pb_calc_s = tcase_create("Calculate s (constant p)");
  tcase_add_test(tc_pb_calc_s, test_pb_calc_s);
  suite_add_tcase(s, tc_pb_calc_s);
  
  /*Test pc_calc_s (variable p) */
  TCase *tc_pb_calc_s_var_p = tcase_create("Calculate s (variable p)");
  tcase_add_test(tc_pb_calc_s_var_p, test_pb_calc_s_var_p);
  suite_add_tcase(s, tc_pb_calc_s_var_p);
  TCase *tc_pb_calc_s_var_p1 = tcase_create("Calculates (variable p) 1");
  tcase_add_test(tc_pb_calc_s_var_p1, test_pb_calc_s_var_p_1);
  suite_add_tcase(s, tc_pb_calc_s_var_p1);

  /*Add test_ps test cases*/
  TCase *tc_ps = tcase_create("Caculate the ps value");
  tcase_add_test(tc_ps, test_ps_ztrue);
  tcase_add_test(tc_ps, test_ps_zfalse);
  tcase_add_test(tc_ps, test_ps_zsum);
  suite_add_tcase(s, tc_ps);
  
  /*Test updating probility method*/
  TCase *tc_pb_update0 = tcase_create("Update probability 0");
  tcase_add_test(tc_pb_update0, test_pb_udpate0);
  TCase *tc_pb_update1 = tcase_create("Update probability 1");
  tcase_add_test(tc_pb_update1, test_pb_udpate1);
  TCase *tc_pb_update2 = tcase_create("Update probability 2");
  tcase_add_test(tc_pb_update2, test_pb_udpate2);
  suite_add_tcase(s, tc_pb_update0);
  suite_add_tcase(s, tc_pb_update1);
  suite_add_tcase(s, tc_pb_update2);
  
  /* Test generalised probability update method */
  TCase *tc_pb_update_var_p = tcase_create("Generalised Update probability");
  tcase_add_test(tc_pb_update_var_p, test_pb_update_var_p);
  suite_add_tcase(s, tc_pb_update_var_p);

  /* Test computing I */
  TCase *tc_I = tcase_create("Test computing I ");
  tcase_add_test(tc_I, test_I);
  suite_add_tcase(s, tc_I);
  
  /* Test computing U */
  TCase *tc_U = tcase_create("Test computing U ");
  tcase_add_test(tc_U, test_U);
  suite_add_tcase(s, tc_U);

  TCase *tc_U_oneloop = tcase_create("Test computing U (one loop) ");
  tcase_add_test(tc_U_oneloop, test_U_oneloop);
  suite_add_tcase(s, tc_U_oneloop);
  
  return s;
}

int
main (void)
{
  int number_failed;
  Suite *s = pb_probability_suite();
  SRunner *sr = srunner_create(s);
  
  /* For debugging test results */
   srunner_set_fork_status (sr, CK_NOFORK);

  srunner_run_all(sr, CK_NORMAL);
  number_failed = srunner_ntests_failed(sr);
  srunner_free(sr);
  return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}


