#include <stdlib.h>
#include <stdio.h>
#include <check.h>

#include "../src/pc_paritycheck.h"
#include "../src/pb_probability.h"
#include "../src/bitarray.h"

START_TEST (test_pc_create_parchecks)
{
  /* create input data */
  unsigned int N = 10;
  unsigned int t = 3;
  /* Feedback poly is 1 + X + X^3 + X^4 */
  unsigned int taps [] = {3,1,4};
  
  /* compute answer */
  pc_paritychecks_t *parchecks = pc_create(N, t, taps);
  
  /* check answer */
  fail_unless(parchecks != NULL, "Null pointer instead on rel normal values");
  fail_unless(pc_numpolys(parchecks) == 2, "Incorrect number of polynomials");
  fail_unless(pc_N(parchecks) == N, "Length of keystream incorrectly set");
 
 
  /* Test first polynomial */
  unsigned int **polys = pc_polys(parchecks);
  unsigned int ans_polys [2] [4] = {{4,1,3,0}, {8,2,6,0}};
  
  
  for (unsigned int i = 0; i < pc_numpolys(parchecks); i++)
    {
      for (unsigned int j = 0; j < t + 1; j++)
	{
	  fail_unless(polys[i][j] == ans_polys[i][j], 
		      "Polynomial %u: term %u is %u instead of %u ", 
		      i, j , polys[i][j], ans_polys[i][j]);
	}
    }
  
  /*clean up*/
  pc_destroy(parchecks);
  
}
END_TEST

START_TEST (test_pc_create_edge)
{
  unsigned int N = 513;
  unsigned int t = 2;
  /* Feedback poly is 1 + X + X^2 */
  unsigned int taps [] = {1,2};
  
  pc_paritychecks_t *parchecks = pc_create(N, t, taps);
  /*
   * Test last polynomial
   * 1 + X^256 + X^512
   */
  unsigned int *poly = pc_polys(parchecks)[pc_numpolys(parchecks) - 1];
  unsigned int ans_poly [] = {512, 256, 0};
  
  for (unsigned int j = 0; j < t + 1; j++)
    {
      fail_unless(poly[j] == ans_poly[j],
  		  "Last Polynomial: term %u is %u instead of %u ",
  		  j , poly[j], ans_poly[j]);
    }

  /*clean up*/
  pc_destroy(parchecks);
  
}
END_TEST

/* 
 * Test creating parity check struct where only one polynomial is generated
 */
START_TEST(test_pc_create_one_poly)
{
  unsigned int N = 6;
  unsigned int taps [] = {2,3}; 
  pc_paritychecks_t *pc =  pc_create(N, 2, taps);
  
  fail_unless(pc_numpolys(pc) == 1);
   
}
END_TEST

START_TEST(test_pc_h)
{
  /* 
   * Create observed keystream 
   *(copied from christine's stream cipher tutorial)
   */
  unsigned int N = 6;
  bit_array_t *z = BitArrayCreate(N);
  BitArraySetAll(z);
  BitArrayClearBit(z, 0);
  BitArrayClearBit(z, 5);
  
  /* Create parchecks object 
   * geffe keystream generator
   * correlation with third LFSR
   * 1 + X^2 + X^3
   */
  unsigned int taps [] = {2,3}; 
  
  pc_paritychecks_t *pc =  pc_create(N, 2, taps);
  /* Results to store answer in */
  unsigned int h [N];

  for (unsigned int i = 0; i < N; i++)
    {
      h[i] = 0;
    }
  
  /* compute result */
  pc_h(pc, z, h);

  /* Test result */
  unsigned int h_p [] = {1,1,1,2,0,1};
  
  for (unsigned int i = 0; i < N; i++)
    {
      fail_unless(h_p[i] == h[i],
		  "Expected bit %u to satisfy %u eqs but instead computed %u",
		  i, h_p[i], h[i]);
    }  
 }
END_TEST

START_TEST(test_pc_h_2)
{
  /* 
   * Create observed keystream 
   *(copied from christine's stream cipher tutorial)
   */
  unsigned int N = 8;
  bit_array_t *z = BitArrayCreate(N);
  BitArraySetAll(z);
  BitArrayClearBit(z, 0);
  BitArrayClearBit(z, 5);
  
  /* Create parchecks object 
   * geffe keystream generator
   * correlation with third LFSR
   * 1 + X^2 + X^3
   */
  unsigned int taps [] = {2,3}; 
  pc_paritychecks_t *pc =  pc_create(N, 2, taps);

  /* Results to store answer in */
  unsigned int h [N];
  for (unsigned int i = 0; i < N; i++)
    {
      h[i] = 0;
    }
  
  /* compute result */
  pc_h(pc, z, h);

  /* Test result */
  unsigned int h_p [] = {2,1,2,2,1,2,1,1};
  
  for (unsigned int i = 0; i < N; i++)
    {
      fail_unless(h_p[i] == h[i],
		  "Expected bit %u to satisfy %u eqs but instead computed %u",
		  i, h_p[i], h[i]);
    }  
 }
END_TEST

/* Test evaluating parity checks and s_i values */
START_TEST (test_pc_parcheck_s)
{
  unsigned int N = 6;
  unsigned int taps [] = {1,2};
  bit_array_t *z = BitArrayCreate(N);
  BitArraySetAll(z);
  BitArrayClearBit(z, 1);
  /* BitArrayClearBit(z, 5); */
  
  double p [] = {0.34 , 0.7, 0.4, 0.9 , 0.99, 0.2};
  
  double **s = (double **)calloc(N, sizeof(double *));
  for (unsigned int i = 0; i < N; i++)
    {
      s[i] = (double *) calloc(PC_NUM_PRODS, sizeof(double));
    }
     
  /*Hope above mem alloc doesn't fail */
  pc_paritychecks_t *pc = pc_create(N, 2, taps);
  
  pc_parcheck_s(pc, z, p, s);
  
  /* yes I worked this out by hand :) */
  double p1 = p[1]*p[2] + (1-p[1])*(1-p[2]);
  double p2 = p[4]*p[2] + (1-p[4])*(1-p[2]);  
  double s_ans [] [PC_NUM_PRODS]  = { 
    {p1, 1 - p1, p2, 1 - p2}
  };
			  
  for (int i = 0; i < 1; i++)
    {
      for (int j = 0; j < 5; j++)
	{
	  fail_unless(dbl_equals(s[i][j],s_ans[i][j]), 
		      "\nFailed on bit %d and parameter %d\nOuput %.10f instead of %.10f", 
		      i, j,
		      s[i][j], s_ans[i][j]);
	}
    }

  /* As a sanity test print out and inspect other values */  
  /* for (unsigned int i = 0; i < N; i++) */
  /*   { */
  /*     printf("%d : %.8f\t%.8f\t%.8f\t%.8f\n", */
  /* 	     i, */
  /* 	     s[i][PC_SATISFIED_PROD], */
  /* 	     s[i][PC_SATISFIED_PROD_COMPLEMENT], */
  /* 	     s[i][PC_UNSATISFIED_PROD], */
  /* 	     s[i][PC_UNSATISFIED_PROD_COMPLEMENT]); */
  /*   }  */
}
END_TEST

/* LFSR test function */
START_TEST(test_pc_check_lfsr_correct)
{
  unsigned int taps [] = {1, 3, 4};
  pc_paritychecks_t *pc = pc_create(8, 3, taps);
  bit_array_t *z = BitArrayCreate(8);
  BitArraySetAll(z);
  BitArrayClearBit(z, 1);
  BitArrayClearBit(z, 4);
  BitArrayClearBit(z, 7);
  
  unsigned char computed = pc_check_lfsr(pc, z);
  fail_unless(computed == 0,
	      "Returned: %u instead of 0",
	      computed);

}
END_TEST

START_TEST(test_pc_check_lfsr_incorrect)
{
  unsigned int taps [] = {1, 3, 4};
  pc_paritychecks_t *pc = pc_create(9, 3, taps);
  bit_array_t *z = BitArrayCreate(9);
  BitArraySetAll(z);
  BitArrayClearBit(z, 1);
  BitArrayClearBit(z, 4);
  BitArrayClearBit(z, 7);
  /* Last bit, a_8,  wrong */
  BitArrayClearBit(z, 8);


  unsigned char computed = pc_check_lfsr(pc, z);
  fail_unless(computed == 1,
	      "Returned: %u instead of 1",
	      computed);
  
}
END_TEST


Suite *
pc_paritycheck_suite(void)
{
  Suite *s = suite_create("Parity Checks");

  TCase *tc_pc_create = tcase_create("Test basic parity check creation");
  tcase_add_test(tc_pc_create, test_pc_create_parchecks);
  suite_add_tcase(s, tc_pc_create);
  
  TCase *tc_pc_create_edge = tcase_create("Test pc_create on edge case");
  tcase_add_test(tc_pc_create_edge, test_pc_create_edge);
  suite_add_tcase(s, tc_pc_create_edge);
  
  TCase *tc_pc_create_one_poly = tcase_create("pc_create generates only one polynomial");
  tcase_add_test(tc_pc_create_one_poly, test_pc_create_one_poly);
  suite_add_tcase(s, tc_pc_create_one_poly);
  
  TCase *tc_pc_h = tcase_create("Test counting relations satisfied");
  tcase_add_test(tc_pc_h, test_pc_h);
  suite_add_tcase(s, tc_pc_h);

  TCase *tc_pc_h_2 = tcase_create("Test counting relations satisfied (2)");
  tcase_add_test(tc_pc_h_2, test_pc_h_2);
  suite_add_tcase(s, tc_pc_h_2);
  
    /* Test evaluating parity checks and computing s */
  TCase *tc_pc_parcheck_s = tcase_create("Test computing parity checks");
  tcase_add_test(tc_pc_parcheck_s, test_pc_parcheck_s);
  suite_add_tcase(s, tc_pc_parcheck_s);
  
  /* Test check if lfsr function */
  TCase *tc_pc_check_lfs_correct = tcase_create("Test correct lfsr seq");
  tcase_add_test(tc_pc_check_lfs_correct, test_pc_check_lfsr_correct);
  TCase *tc_pc_check_lfs_incorrect = tcase_create("Test incorrect lfsr seq");
  tcase_add_test(tc_pc_check_lfs_incorrect, test_pc_check_lfsr_incorrect);
  suite_add_tcase(s, tc_pc_check_lfs_correct);
  suite_add_tcase(s, tc_pc_check_lfs_incorrect);
  return s;
}


int 
main (void)
{
  int number_failed;
  Suite *s = pc_paritycheck_suite();
  SRunner *sr = srunner_create(s);

  /* For easier debugging of tests */
  srunner_set_fork_status (sr, CK_NOFORK);

  srunner_run_all(sr, CK_NORMAL);
  number_failed = srunner_ntests_failed(sr);
  srunner_free(sr);
  return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
